@@ -1,9 +1,8 @@
-from .addresses import *
 from .constants import OUTPUT_ROOT_PROOF_VERSION, CHALLENGE_PERIOD_MAINNET, CHALLENGE_PERIOD_TESTNET
 
-from .types import MessageStatus
-from .utils import get_provider, get_account, to_low_level_message, make_state_trie_proof, hash_message_hash
-from .contracts import L2ToL1MessagePasser, L2OutputOracle, OptimismPortal, CrossChainMessengerContract
+from .types import MessageStatus, OutputRootProof, BedrockMessageProof
+from .utils import get_provider, get_account, to_low_level_message, make_state_trie_proof, hash_message_hash, read_addresses, load_abi
+from .contracts import L2ToL1MessagePasser, L2OutputOracle, OptimismPortal, CrossChainMessengerContract, StandardBridge
 
 class CrossChainMessenger():
 
@@ -35,15 +34,65 @@ class CrossChainMessenger():
         self.l1_cross_chain_messenger = CrossChainMessengerContract(account_l1, chain_id_l1, chain_id_l2, provider=provider_l1, network=network)
         self.l2_cross_chain_messenger = CrossChainMessengerContract(account_l2, chain_id_l2, chain_id_l1, provider=provider_l2, network=network)
 
+        self.l1_bridge = StandardBridge(account_l1, from_chain_id=chain_id_l1, to_chain_id=chain_id_l2, provider=provider_l1, network=network)
+        self.l2_bridge = StandardBridge(account_l2, from_chain_id=chain_id_l2, to_chain_id=chain_id_l1, provider=provider_l2, network=network)
+
     def deposit_eth(self, value):
         
-        return self.l1_cross_chain_messenger.send_message(self.account_l1.address, b"", 0, value)
+        return self.l1_cross_chain_messenger.send_message(self.account_l2.address, b"", 0, value)
+    
+    def deposit_erc20(self, token_address_l1, token_address_l2, value):
+        
+        if not self._supports_token_pair(token_address_l1, token_address_l2):
+            raise Exception("Token pair not supported")
+        
+        return self.l1_bridge.deposit_erc20(token_address_l1, token_address_l2, value, 0, b"")
+    
+    def deposit_erc20_to(self, token_address_l1, token_address_l2, to, value):
+        
+        if not self._supports_token_pair(token_address_l1, token_address_l2):
+            raise Exception("Token pair not supported")
+        
+        return self.l1_bridge.deposit_erc20_to(token_address_l1, token_address_l2, to, value, 0, b"")
+
+    def approve_erc20(self, token_address_l1, token_address_l2, value):
+
+        if not self._supports_token_pair(token_address_l1, token_address_l2):
+            raise Exception("Token pair not supported")
+
+        token_contract_l1 = self.provider_l1.eth.contract(address=token_address_l1, abi=load_abi("ERC20"))
+
+        approve_tx = token_contract_l1.functions.approve(self.l1_bridge.address, value).build_transaction({
+            "from": self.account_l1.address,
+            "nonce": self.provider_l1.eth.get_transaction_count(self.account_l1.address)
+        })
+
+        signed_txn = self.provider_l1.eth.account.sign_transaction(approve_tx, self.account_l1.key)
+        txn_hash = self.provider_l1.eth.send_raw_transaction(signed_txn.rawTransaction)
+        txn_receipt = self.provider_l1.eth.wait_for_transaction_receipt(txn_hash)
+
+        return txn_hash.hex(), txn_receipt
+
 
     def withdraw_eth(self, value):
         
         l2_to_l1_message_passer = L2ToL1MessagePasser(self.account_l2, provider=self.provider_l2, network=self.network)
         
-        return l2_to_l1_message_passer.initiate_withdrawl(self.account_l2.address, 0, b"", value)
+        return l2_to_l1_message_passer.initiate_withdrawl(self.account_l1.address, 0, b"", value)
+    
+    def withdraw_erc20(self, token_address_l1, token_address_l2, value):
+
+        if not self._supports_token_pair(token_address_l1, token_address_l2):
+            raise Exception("Token pair not supported")
+        
+        return self.l2_bridge.bridge_erc20(token_address_l1, token_address_l2, value, 0, b"")
+    
+    def withdraw_erc20_to(self, token_address_l1, token_address_l2, to, value):
+
+        if not self._supports_token_pair(token_address_l1, token_address_l2):
+            raise Exception("Token pair not supported")
+            
+        return self.l2_bridge.bridge_erc20_to(token_address_l1, token_address_l2, to, value, 0, b"")
     
     def prove_message(self, l2_txn_hash):
         
@@ -51,10 +100,10 @@ class CrossChainMessenger():
         l2_txn_receipt = self.provider_l2.eth.get_transaction_receipt(l2_txn_hash)
 
         withdrawl_message, withrawl_message_hash = to_low_level_message(l2_txn, l2_txn_receipt)
-        proof = self.get_bedrock_message_proof(l2_txn, withrawl_message_hash)
+        message_proof = self.get_bedrock_message_proof(l2_txn, withrawl_message_hash)
 
         optimism_portal = OptimismPortal(self.account_l1, provider=self.provider_l1, network=self.network)
-        return optimism_portal.prove_withdrawl_transaction(tuple(withdrawl_message.values()), proof["l2OutputIndex"], tuple(proof["outputRootProof"].values()), tuple(proof["withdrawalProof"]))
+        return optimism_portal.prove_withdrawl_transaction(withdrawl_message.values(), message_proof.get_l2_output_index(), message_proof.get_output_root_proof(), message_proof.get_withdrawl_proof())
     
     def finalize_message(self, l2_txn_hash):
 
@@ -63,7 +112,7 @@ class CrossChainMessenger():
 
         withdrawl_message, withrawl_message_hash = to_low_level_message(l2_txn, l2_txn_receipt)
         
-        optimism_portal = OptimismPortal(self.account_l1, network=self.network)
+        optimism_portal = OptimismPortal(self.account_l1, provider=self.provider_l1, network=self.network)
         return optimism_portal.finalize_withdrawl_transaction(tuple(withdrawl_message.values()))
 
 
@@ -78,20 +127,20 @@ class CrossChainMessenger():
 
         message_slot = hash_message_hash(withdrawl_hash)
 
-        state_trie_proof = make_state_trie_proof(self.provider_l2, l2_block_number, L2_TO_L1_MESSAGE_PASSER, message_slot)
+        state_trie_proof = make_state_trie_proof(self.provider_l2, l2_block_number, read_addresses("l2")["l2_" + self.network]["L2_TO_L1_MESSAGE_PASSER"], message_slot)
 
         block = self.provider_l2.eth.get_block(l2_block_number)
+
+        output_root_proof = OutputRootProof(version=OUTPUT_ROOT_PROOF_VERSION,
+                                            state_root=block.stateRoot.hex(),
+                                            message_passer_storage_root=state_trie_proof.storage_root.hex(),
+                                            latest_blockhash=block.hash.hex())
         
-        return {
-            "outputRootProof": {
-                "version": OUTPUT_ROOT_PROOF_VERSION,
-                "stateRoot": block.stateRoot.hex(),
-                "messagePasserStorageRoot": state_trie_proof["storage_root"].hex(),
-                "latestBlockhash": block.hash.hex(),
-            },
-            "withdrawalProof": [el.hex() for el in state_trie_proof["storage_proof"]],
-            "l2OutputIndex": latest_l2_output_index,
-        }
+        bedrock_message_proof = BedrockMessageProof(output_root_proof=output_root_proof,
+                                                    withdrawl_proof=[el.hex() for el in state_trie_proof.storage_proof],
+                                                    l2_output_index=latest_l2_output_index)
+        
+        return bedrock_message_proof
     
     def get_message_status(self, txn_hash):
 
@@ -170,4 +219,14 @@ class CrossChainMessenger():
         raise NotImplementedError
     
     def get_state_root_batch_by_transaction_index(self):
-        raise NotImplementedError
\ No newline at end of file
+        raise NotImplementedError
+    
+    def _supports_token_pair(self, token_address_l1, token_address_l2):
+        
+        token_contract_l2 = self.provider_l2.eth.contract(address=token_address_l2, abi=load_abi("OPTIMISM_ERC20"))
+        try:
+            l1_token = token_contract_l2.functions.l1Token().call()
+        except:
+            raise Exception(f"Token Contract {token_address_l2} is not an Optimism ERC20")
+        
+        return l1_token == token_address_l1
\ No newline at end of file
